# Create a python project for an SPR extension module.
# import Simple_Process_REPL.appstate as A
import os
import pkgutil
import regex as re
import logging

logger = logging.getLogger()

root = "make-ext"


def get_pkg_file(fname):
    return pkgutil.get_data("Simple_Process_REPL", fname)


def insert_name(name, lines):
    """Look through a list of lines looking for %s formats. if found,
    format it with the name."""
    res = []
    for line in lines.split("\n"):
        line = re.sub("%s", name, line)
        line = re.sub("%yaml", "%s", line)
        res += [line + "\n"]
    return res


def write_lines_2_file(fname, lines):
    """Open a file and write the lines to it."""
    with open(fname, "w") as f:
        f.writelines(lines)
    f.close()


def new_spr_extension_project(pathname):
    """Create a Python project to create an SPR extension module.
    This creates a skeleton python project that is
    compatible with pip.

    It includes everything needed to create a new extension
    module that can be installed with pip and imported by SPR.
    """

    logger.info("Creating Python Project: %s" % pathname)

    if os.path.exists(pathname):
        logger.error("Path: %s, exists" % pathname)
        return
    else:
        os.mkdir(pathname)

    name = os.path.split(pathname)[1]
    module_path = os.path.join(pathname, name)

    setup = insert_name(name, get_pkg_file("setup.txt").decode("utf-8"))
    py = insert_name(name, get_pkg_file("python.txt").decode("utf-8"))
    spr = insert_name(name, get_pkg_file("spr.txt").decode("utf-8"))
    yaml = insert_name(name, get_pkg_file("yaml.txt").decode("utf-8"))
    readme = insert_name(name, get_pkg_file("readme.txt").decode("utf-8"))

    os.mkdir(module_path)

    if setup:
        write_lines_2_file(os.path.join(pathname, "setup.py"), setup)

    if readme:
        write_lines_2_file(os.path.join(pathname, "README.md"), readme)

    write_lines_2_file(
        os.path.join(module_path, "__init__.py"), ['__version__ = "0.0.1"']
    )

    if py:
        write_lines_2_file(os.path.join(module_path, "core.py"), py)

    if spr:
        write_lines_2_file(os.path.join(module_path, "core.spr"), spr)
    if yaml:
        write_lines_2_file(os.path.join(module_path, "core.yaml"), yaml)



import pkgutil
from platform import system as platform
from Simple_Process_REPL.options import create_parser
import Simple_Process_REPL.logs as logs
import os
import logging
import Simple_Process_REPL.repl as r
import Simple_Process_REPL.utils as u
from Simple_Process_REPL.dialog import hello
import yaml

logger = logging.getLogger()

# format and fill in as you wish.
HelpText = """
appstate: - Manage SPR's yaml datastore.  -

Everything needed to load, save, copy, set, and merge data
in the yaml datastore. Yaml files and merging etc.

Appstate is the Yaml Datastore.

When writing python code to interact with SPR the appstate functions
get_in, set_in, get_in_config, and get_in_device are of primary
use.

Within SPR code, show, set, and set-from are of primary use.

"""


def help():
    """Additional SPR help For the appstate Module."""
    print(HelpText)


# yaml datastore, which will contain merged data from the application layer.
AS = {
    "config": {},
    "args": {"commands": {}},
    "defaults": {
        "config_file": "PBRConfig.yaml",
        "loglevel": "info",
        "logfile": "PBR.log",
    },
    "platform": "",
}

with_stack = []


def _ls_with():
    path = with_stack[-1]["path"]
    print(path)
    print("-------------------------")
    if path[0] == "/":
        if len(path) == 1:
            keys = get_keys_in()
        else:
            vv = path[1:].split("/")
            keys = get_keys_in(*vv)

    if keys:
        for k in keys:
            print("    %-30s" % k)
    else:
        print("   Is Empty.")


def _with(wpath=None, wcommand=None, destpath=None):
    """
    Show the current 'With' path or if a path is given,
    Push a Yaml datastore path onto the 'with' stack.

    If it is not yet there, it will appear when someone sets something.

    If a command is given, execute the command and pop.
    if a destination is given pop the last result to that path.
    """
    global with_stack
    #    logger.info("_with %s" % path)
    #    logger.info("_stack %s" % with_stack)

    if wpath is None:
        _ls_with()
        return

    if wpath[0] != "/":
        wpath = "/" + wpath

    # Value vectors, paths, I need to get these
    # symbols straightened out a bit. Seems like
    # should be keeping withs, in the app state too.
    # they seem a little stupid at the moment.
    # basically a manual path stack.
    vv = _get_vv_from_path(wpath)
    try:
        vv.remove("")
    except Exception:
        pass

    if vv is None:
        vv = []

    with_stack.append(
        {
            "path": wpath,
            "vv": vv,
        }
    )
    if wcommand:
        r.eval_cmd(wcommand)
        if destpath is not None:
            pop("results", destpath)
        pop("/_with_")
    else:
        try:
            _ls_with()
        except Exception:
            pass


def _print_stack():
    """Print the stack of 'withs/scopes'.
    Show the contents of the top entry
    and the entries of the stack which are below it."""
    global with_stack
    # print(reversed(with_stack))
    # pretty print the dict at path.
    for p in reversed(with_stack):
        print(p["path"])


def _get_with_path():
    """Return the current 'With' Vector as a path."""
    return with_stack[-1]["path"]


def _get_with_vv():
    """Return the current 'With' Vector."""
    return with_stack[-1]["vv"]


def _full_with_path(path):
    """takes a path and prepends the with path."""
    withpath = _get_with_path()

    if withpath != "/" and path[0] != "/":
        path = "/" + path
    return _get_with_path() + path


# def set_with(set_path, fromv):
#     """Takes a path or path symbol and a value
#     or 2 paths. Prepends the path of with and calls set.
#     """
#     set_path = _full_with_path(set_path)
#     # logger.debug("set-with %s" % set_path)
#     set(set_path, fromv)


def get_in_with(path):
    vv = _get_with_vv() + _get_vv_from_path(path)
    return get_in(vv)


def select_keys(m, keys):
    """Given a map and list of keys,
    return a map of with those keys from the map.
    """
    d = None
    try:
        d = {k: m[k] for k in keys}
    except Exception:
        pass
    return d


def select_with(keys):
    """Given a list of keys,
    return a map of with those keys from the with map.
    """
    return select_keys(get_with(), keys)


def get_with():
    """Return the current with map."""
    return get_in(_get_with_vv())


def _set_(d):
    """Merge in a new dict, like the device dictionary, into
    the yaml datastore.
    """
    global AS
    AS = u.merge(AS, d)


def set_in(*keys):
    """Takes a list of keys ending with the value to assign
    into the Yaml Datastore dictionary tree."""
    global AS
    AS = u.merge(AS, u.make_dict(*keys))


def _get_vv_from_path(path):
    if path[0] == "/":
        path = path[1:]
    return path.split("/")


def get_from_path(path):
    vv = _get_vv_from_path(path)
    return get_in(vv)


def clear_path(path):
    """clear the value at path."""
    global AS
    set_keys = get_vv(path)
    set_keys += [""]
    AS = u.merge(AS, u.make_dict(set_keys))


def pop(path, destination=None):
    """if the path is a list , pop the last value,
    if it is not a list, clear it.
    If destination is set, place the value popped there.
    """
    if path[0] != "/":
        value = get_in_with(path)
    else:
        value = get_from_path(path)

    if isinstance(value, list):
        if len(value):
            v = value.pop()
        else:
            v = None

    else:
        clear_path(path)
        v = value

    if destination is not None:
        # logger.info(destination)
        set(destination, v)


def push(set_path, fromv):
    """Takes a path or path symbol and a value
    or 2 paths.

    Honors 'with'.

    Values beginning with / will be
    treated a path, otherwise as a symbol/value.

    pushes the value onto the list at set_path.
    If set_path is not a list, it will be turned into one.
    """
    global AS

    set_keys = _get_vv_from_path(_full_with_path(set_path)[1:])
    val = get_fromv(fromv)

    try:
        dest = get_in_with(set_path)
    except Exception:
        dest = None

    if dest:
        if not isinstance(dest, list):
            dest = [dest]
            dest += [val]
    else:
        dest = [val]

    # logger.info("push: %s" % dest)
    # logger.info("keys: %s" % set_keys)

    set_keys += [dest]
    AS = u.merge(AS, u.make_dict(set_keys))


def get_fromv(fromv):
    """Get the value from there, if it's a there.
    If its raw value is a list, then it's a string from the parser.
    if it's a path, then get the value."""
    if isinstance(fromv, int) or isinstance(fromv, float):
        fromv = fromv

    elif isinstance(fromv, str):
        # path = r.isa_path(fromv)
        # if path:
        #     fromv = _get_vv_from_path(path)
        # else:

        if fromv[0] == "/":
            fromv = get_from_path(fromv[1:])

    elif isinstance(fromv, list):
        res = None
        for x in fromv:
            if res is None:
                res = str(x)
            else:
                res = res + " " + str(x)
                fromv = res

    elif fromv and len(fromv) == 1:
        fromv = fromv  # [0]

    return fromv


def get_vv(set_path):
    """return the value vector for a path."""
    set_keys = _get_vv_from_path(set_path)
    path = r.isa_path(set_path)
    if path:
        set_keys = _get_vv_from_path(path)
    else:
        set_keys = _get_vv_from_path(set_path)

    return set_keys


def _merge(fromv, set_path=None):
    """merge a dictionary into the current with"""
    fromv = get_fromv(fromv)

    if set_path is None:
        set_keys = _get_with_vv()

    elif set_path[0] != "/":
        set_path = _full_with_path(set_path)
        logger.debug("set-with %s" % set_path)
        set_keys = get_vv(set_path)

    set_keys += [fromv]
    AS = u.merge(AS, u.make_dict(set_keys))
    u.merge(AS, d)


def set(set_path, fromv):
    """Takes a path or path symbol and a value
    or 2 paths.

    Set path can be absolute, or relative to the with.
    If set_path does not begin with / then it
    will be searched from the 'with' root.

    Values beginning with / will be
    treated as a path, otherwise as a symbol/value.
    """
    global AS
    # with local and absolute paths.
    if set_path is None:
        set_keys = _get_with_vv()

    elif set_path[0] != "/":
        set_path = _full_with_path(set_path)
        logger.debug("set-with %s" % set_path)
        set_keys = get_vv(set_path)

    fromv = get_fromv(fromv)

    set_keys += [fromv]
    AS = u.merge(AS, u.make_dict(set_keys))


def get(path):
    """Get a data value from a path in the datastore.
    Without a leading / the path is considered relative
    to the current 'with' path.
    """
    vv = _get_vv_from_path(path)

    if path[0] != "/":
        vv = _get_with_vv() + vv

    return get_in(vv)


def get_in(keys):
    """Get something out of the yaml datastore."""
    global AS
    return _get_in(AS, keys)


def get_keys_in(*keys):
    """Get just a list of keys from the thing at the value vector."""
    thing = get_in(keys)
    if isinstance(thing, list):
        return thing
    elif thing:
        return thing.keys()
    return None


def get_vals_in(path, *keys):
    """Get a list of values from a value vector in the yaml datastore.
    Vector should be a vector of keys, and keys should be a list of keys. :-/

    this, that, it = get_vals_in(["foo", "bar"], ["this" "that" "it"]).

    This is of primary use in the creation '-with' commands.
    """
    res = []
    d = get_from_path(path)
    for k in keys[0]:
        try:
            res += [d[k]]
        except Exception:
            res += [None]
    return res


def _get_in(dict_tree, keys):
    """
    Retrieve a value from a dictionary tree, using a key list
    Returns:
       The value found at the given key path, or `None` if
       any of the keys in the path is not found.
    """

    # logger.info("_get_in")
    # logger.info(dict_tree.keys())
    # logger.info(keys)
    # logger.info("_get_in")

    try:
        for key in keys:
            # logger.info("key %s" % key)
            dict_tree = dict_tree[key]

        return dict_tree

    except KeyError:
        return None


# could have been a partial.
def get_in_config(keys):
    "Get stuff from the config, takes a list of keys."
    return _get_in(AS["config"], keys)


def get_in_device(key):
    "Get to the device info, easier to read."
    return _get_in(AS["device"], [key])


def show(pathname=None):
    """Show a sub-tree in the Yaml Datastore with a path.

    If no path is given, the current _with_ is shown. if
    the current _with_ is '/', it is listed instead of
    shown, as showing that is rarely what we would want.

    If you really want to 'show' the entire tree it has to
    be explicit.  `show /`

    example: show /device

    This will display the yaml datastore tree from the device node on.
    """
    qqc = None

    # point at with if we got Nothing.
    if pathname is None:
        pathname = _get_with_path()
        # We really don't want to 'show' that on accident.
        if pathname == "/":
            pathname = None
            _ls_with()

    if pathname == "/":
        # remove _Root_ from showing unless asked.
        qqc = AS | {"_Root_": None}

    elif pathname:
        logger.info(pathname)
        # so we don't have to type the first / in the path.
        if pathname[0] == "/":
            pathname = pathname[1:]
        vv = pathname.split("/")
        qqc = get_in(vv)

    if qqc:
        logger.info(yaml.dump(qqc))


def archive_log(new_name):
    "Move/rename the current logfile to the filename given."
    os.rename(get_in_config(["files", "logfile"]), new_name)


def sync_functions():
    "Sync user functions from the interpreter to the config."
    funcs = r.get_user_functions()
    AS["config"]["exec"]["functions"] = funcs


def islinux():
    """Check if the platform is linux."""
    return "Linux" == AS["platform"]


def reset_device():
    "Start fresh with empty device values."
    new_device = {}
    id = get_in_device("id")

    for k, v in AS["device"]:
        new_device[k] = ""

    new_device["last_id"] = id
    AS["device"] = new_device


def eval_default_process():
    "Run the autoexec process"
    autoexec = get_in_config(["autoexec"])
    if autoexec is not None:
        try:
            r.eval_cmd(autoexec)
        except Exception as e:
            logger.error(e)
            raise Exception(e)
    else:
        hello()


def merge_yaml_with(y):
    """Merge a yaml data structure into the yaml datastore."""
    logger.debug("Merge yaml-\n %s" % y)
    root = _get_with_vv()
    y = [yaml.load(y, Loader=yaml.SafeLoader)]
    # logger.info("merge: %s" % root)
    y = root + y
    d = u.make_dict(y)

    # logger.info(d)
    u.merge(AS, d)


def merge_yaml(y):
    """Merge a yaml data structure into the yaml datastore."""
    logger.debug("Merge yaml-\n %s" % y)
    u.merge(AS, yaml.load(y, Loader=yaml.SafeLoader))


def load_yaml(yaml_file):
    """Load a yaml file into the yaml datastore"""
    merge_yaml(u.load_yaml_file(yaml_file))


def load_functions():
    """Give the functions from the configuration to the repl.
    by adding them to the symbol table."""
    # add in the user functions from the config file.

    fns = get_in_config(["exec", "functions"])

    if fns is not None:
        for k, v in fns.items():
            r.def_symbol(k, v["doc"], v["fn"])


def load_defaults(state_init, pkgname=None, yamlname=None):
    global AS

    bc = load_base_config()
    if bc:
        AS["config"] = u.merge(AS["config"], bc)
        # AS |= state_init  #### destructive...
    AS = u.merge(AS, state_init)
    if pkgname is None:
        return AS
    AS["config"] = u.merge(AS["config"], u.load_pkg_yaml(pkgname, yamlname))
    return AS


def merge_pkg_yaml(package, yamlname):
    """load a yaml file from a package into the yaml datastore."""
    global AS
    AS = u.merge(AS, u.load_pkg_yaml(package, yamlname))


def load_base_config():
    """load the default configuration."""
    return u.load_pkg_yaml(__name__, "SPR-defaults.yaml")


def load_pkg_resource_to(package, filename, *keys):
    """load a python package resource file and place the contents
    at the value vector given.

    example: load-pkg-resource-to Simple_Process_REPL README.md readme md

    Will place the contents of the README.md file into the value vector
    readme/md.
    """
    res = u.load_pkg_resource(package, filename)
    set_in(keys[0] + [res])


def load_pkg_resource(package, filename):
    """load a python package resource file.
    Find a resource file in a python module
    and return it.
    """
    return u.load_pkg_resource(package, filename)


def load_pkg_resource_with(path):
    """load a python package resource file
    using the values found in 'package' and 'filename' located at
    the path.

    The contents will be placed into 'contents' at the path given.

    For this example there is a structure in the yaml datastore which
    looks like this. In this case the value vector is simply 'readme'.
    Contents is included for clarity, but will be created if not there.

    readme:
         package: "Simple_Process_REPL"
         filename: "README.md"
         content: ""

    example: load-pkg-resource-with readme

    Will place the contents of the README.md file into the value vector
    readme/md.

    In SPR:

    with readme
    '
    package: Simple_Process_REPL
    filename: README.md


    load-pkg-resources-with readme

    """
    pkgname, filename = get_vals_in(path, ["package", "filename"])
    res = u.load_pkg_resource(pkgname, filename)
    keys = _get_vv_from_path(path)
    set_in(keys + ["content", res])


def save_config(filename):
    "Sync the functions from the interpreter and save the configuration."
    sync_functions()
    u.save_yaml_file(filename, AS["config"])


def load_config(filename):
    """load a yaml file into the application's
    configuration dictionary.
    """
    AS["config"] = u.load_yaml_file(filename)


def load_configs():
    global AS
    cli_config = get_in(["args", "config_file"])
    defaults = get_in(["defaults", "config_file"])
    if cli_config is not None:
        y = u.load_yaml_file(cli_config)
        if y is not None:
            AS["config"] = u.merge(AS["config"], y)
    elif defaults is not None:
        y = u.load_yaml_file(defaults)
        if y is not None:
            AS["config"] = u.merge(AS["config"], y)


def init(parser, logger):
    """
    Parse the cli parameters,
    load the default config or the configuration given,
    start logging,
    initialize the symbol tables for the interpreter.
    Finally, run in whatever mode we were told.
    """
    global AS

    if parser is None:
        parser = create_parser(get_in(["defaults"]))

    AS["args"] = vars(parser.parse_args())

    _with("/")

    load_configs()

    logs.add_file_handler(
        logger,
        get_in_config(["files", "loglevel"]),
        get_in_config(["files", "logfile"]),
    )
    set_in(["platform", platform()])
    set_in(["_with_", with_stack])

    # load functions from the config into the interpreter.
    load_functions()



import logging
import regex as re
import Simple_Process_REPL.appstate as A
import Simple_Process_REPL.utils as u

logger = logging.getLogger()

yaml = u.dump_pkg_yaml("Simple_Process_REPL", "dialog.yaml")

HelpText = (
    """
cli: An interface to the commandline using python input().

This module provides a few cli prompts, and messages.
it is a partial match to the interface presented in the dialog module.

There is currently no stateful data, although that should change.

The cli uses the same parts of the configuration as the dialog module.

%s

"""
    % yaml
)


def help():
    """Additional SPR specific Help for the cli Module."""
    print(HelpText)


def hello():
    "Just in case we don't know what to do."
    msg(A.get_in_config(["dialogs", "hellomsg"]))


def radiolist(msg, choices):
    pass


def select_choice(msg, choices):
    pass
    # return choice


def print_command_menu():
    """Give a menu of possible print commands for the current platform"""
    p_dict = A.get_in_config(["print_commands", A.get_in(["platform"])])
    p_p_cmds = list(p_dict.items())

    rlist = [(key, "") for key, cmd in p_p_cmds]
    choice = select_choice("Which print command to use ?", rlist)
    return choice, p_dict[choice]


def yes_no(msg):
    "Yes or No message, returns True or False."
    if re.match("^[yY]?$", input(message)):
        return True
    return False


def continue_to_next():
    "Continue to next. raise exception if not Y/y."
    if yes_no(A.get_in_config(["dialogs", "continue_to_next"])):
        return True
    raise Exception("Answer was No")


def msg(msg):
    """Display a message on the cli and wait for input."""

    if msg[0] == "/":
        msg = A.get(msg)

    print(msg)
    input("Press any key to continue;")



import os
from dialog import Dialog
import logging

import regex as re
import Simple_Process_REPL.appstate as A
import Simple_Process_REPL.bar_qr as bq
import Simple_Process_REPL.utils as u

logger = logging.getLogger()

# set up the dialog interface
d = Dialog(dialog="dialog")
inputbox = d.inputbox

yaml = u.dump_pkg_yaml("Simple_Process_REPL", "dialog.yaml")
# format and fill in as you wish.
HelpText = (
    """
dialog: An interface to the python dialog library.

This module provides dialog windows for various purposes.

There is currently no stateful data.

    %s

dialog is an old curses library written in C used for simple dialogs.
"""
    % yaml
)


def help():
    print(HelpText)


def hello():
    "Just in case we don't know what to do."
    msg(A.get_in_config(["dialogs", "hellomsg"]))


def input_string(msg):
    """Get a string input"""
    code, res = inputbox(
        msg,
        title=A.get_in_config(["dialogs", "title"]),
        height=10,
        width=50,
    )
    return res


def input_count(msg):
    """Give an input dialog that insists on an integer input.
    A range box could work, but seemed cumbersome."""
    while True:
        code, res = inputbox(
            msg,
            title=A.get_in_config(["dialogs", "title"]),
            height=10,
            width=50,
        )
        try:
            res = int(res)
        except Exception:
            msg("count must be an integer")
            continue
        break
    return res


def _input_string(msg, regex=None, correct=None, must=None, confirmation=True):
    """
    Give an input dialog which will check the entered value against
    a regular expression.

    Allows for the elimination of the confirmation
    dialog and setting of confirmation and error messages.

    If the result matches the regular expression,
    and a 'correct' message is provided,
    Then there will be a confirmation dialog provided.

    If incorrect, the message provided by must or in
    /config/dialogs/must is displayed along with the regular
    expression which failed.

    An example of yaml to hold a regex.
    This one is for 8 digits.
    regex : r"^\d{8}$"

    """
    if regex is None:
        return input_string(msg)

    if must is None:
        must = (A.get_in_config(["dialogs", "must"]),)

    if correct is None:
        correct = A.get_in_config(["dialogs", "correct"])

    while True:
        res = input_string(msg)
        if re.match(regex, res):
            if confirmation:
                yno_msg = "%s : %s" % (correct, res)
                if yes_no(yno_msg):
                    break
            else:
                break
        else:
            if must is not None:
                msg("%s : %s" % ("must", regex))

    return res


def dialog_print(fname):
    """Dialog to ask which print command to use."""
    cmd_name, print_command = print_command_menu()
    command = print_command % fname
    return cmd_name, command


def dialog_print_loop(fname):
    """Dialog to ask which print command to use and how many times to print it."""
    cmd_name, command = dialog_print(fname)
    count = input_count("How many to Print ?")
    return cmd_name, command, count


def msg(msg):
    """Display a simple message box, with either
    the msg or the text located at msg if it begins with a /..
    """
    if msg[0] == "/":
        msg = A.get(msg)

    d.msgbox(
        msg,
        title=A.get_in_config(["dialogs", "title"]),
        height=10,
        width=50,
    )
    os.system("clear")


def yes_no(msg):
    """Display a yesno dialog,
    If msg is a path, use the contents of path.

    return True or False."
    """

    if msg[0] == "/":
        msg = A.get(msg)

    response = d.yesno(
        msg,
        title=A.get_in_config(["dialogs", "title"]),
        height=10,
        width=50,
    )
    os.system("clear")
    if response == "ok":
        return True
    else:
        return False


def radiolist(msg, choices):
    if msg[0] == "/":
        msg = A.get(msg)

    (code, tag) = d.radiolist(
        msg,
        width=50,
        title=A.get_in_config(["dialogs", "title"]),
        choices=choices,
    )
    return tag


def select_choice(msg, choices):
    if msg[0] == "/":
        msg = A.get(msg)

    logger.info("select choice: %s", choices)
    code, choice = d.menu(
        msg,
        title=A.get_in_config(["dialogs", "title"]),
        choices=choices,
        height=50,
        width=50,
    )
    os.system("clear")
    return choice


def print_command_menu():
    """Give a menu of possible print commands for the current platform"""
    p_dict = A.get_in_config(["print_commands", A.get_in(["platform"])])
    p_p_cmds = list(p_dict.items())

    rlist = [(key, "") for key, cmd in p_p_cmds]
    choice = select_choice("Which print command to use ?", rlist)
    return choice, p_dict[choice]


def BC_or_QR_menu():
    return select_choice(
        "Which would you like to print?",
        [
            ("Bar Code", ""),
            ("QR Code", ""),
        ],
    )


def BC_or_QR():
    """Dialog for Bar code or QR, normalized to Codetype."""
    label_type = BC_or_QR_menu()
    if label_type == "Bar Code":
        return BarCodeType
    elif label_type == "QR Code":
        return QRCodeType


def save_bcqr():
    """Ask which; Barcode or Qr code, then generate and save the png,
    return the filename."""
    codetype = BC_or_QR()
    bq.get_bcqr(codetype)
    return bq.save_bcqr(codetype)


def print_bcqr():
    """Dialogs to generate, save, and print any number of the current barQR
    value as a bar or QR code."""
    fn = bq.print_bcqr()
    print_file_loop(fn)


# has nothing to do with printing really.... Just the messages. -- Refactor.
def _print_file(fname, cmd_name, command, count=1):
    """Internal use. Display print messages and loop or not over a system command."""
    if count > 1:
        logger.info("Printing file %s, %d times, to %s" % (fname, count, cmd_name))
    else:
        logger.info("Printing file %s to %s" % (fname, cmd_name))

    if count > 1 and ynbox(
        "You are ready to print %d times to %s?" % (count, cmd_name)
    ):
        for i in range(0, count):
            os.system(command)
    else:
        os.system(command)


def print_file(filename):
    """Print a filename with a series of dialog prompts."""
    name, cmd = dialog_print(filename)
    _print_file(filename, name, cmd, 1)


def print_file_loop(filename):
    """Print a filename A number of times with a series of dialog prompts."""
    name, cmd, count = dialog_print_loop(filename)
    _print_file(filename, name, cmd, count)


def continue_to_next():
    "Do another one? Dialog. returns True/False"
    if yes_no(A.get_in_config(["dialogs", "start_again"])):
        logger.info("exiting")
        return False
    return True



import Simple_Process_REPL.dialog as D
import logging
import regex as re

import Simple_Process_REPL.appstate as A
import Simple_Process_REPL.utils as u
import Simple_Process_REPL.subcmd as S

logger = logging.getLogger()

yaml = u.dump_pkg_yaml("Simple_Process_REPL", "network.yaml")

# format and fill in as you wish.
HelpText = (
    """
network: - Network type stuff Currently only wifi for linux.  -

Uses dialogs and nmcli, network-manager, to establish a wifi connection.

network uses these parts of the Application state.

%s

connect-wifi always checks to see if there is a connection. If there is,
it simply remembers it.

The first time, on a fresh system, if network manager has never connected,
connect-wifi will search for networks and present a menu of choices.
This usually only happens once and only if the computer has never connected
to the wifi in that location.

At some point there may be some ssh type commands here. Currently none of them
do anything.

"""
    % yaml
)


def help():
    """Additional SPR specific Help for the Network Module."""
    logger.info("network help")
    print(HelpText)


def check_connection():
    "Check the wifi connection with Network manager"
    if A.islinux():
        logger.debug("check connection!")
        res = S.do_cmd(["nmcli", "device"])
        logger.debug(res)
        if len(re.findall("connected", res)):
            A.set_in(["network", "wifi-connected", True])
        else:
            A.set_in(["network", "wifi-connected", False])
            return False
        return True
    return False


def get_SSIDS():
    # get first column.
    cmd = ["nmcli", "connection", "show"]
    res = S.do_cmd(cmd)
    ssids = []
    for line in res.split("\n"):
        ssids.append(line.split(" ")[0])

    return ssids


def connect_SSID(ssid):
    """Connect the wifi to the SSID given."""
    cmd = ["nmcli", "device", "wifi", "connect", ssid]
    res = S.do_cmd(cmd)
    logger.info(res)


def connect_wifi():
    """
    Connect the wifi, get the SSID's, give a selection dialog to choose,
    then connect to the chosen SSID. Only implemented for linux for use
    with network manager.
    """
    if A.islinux():
        if not check_connection():
            ssids = get_SSIDS()
            menuitems = []
            for ssid in ssids:
                menuitems.append([ssid, ssid])
            ssid = D.select_choice("Choose an SSID to Connect.", menuitems)
            if ssid is not None:
                logger.info("Connecting to %s" % ssid)
                connect_SSID(ssid)


def confirm_tunnel():
    "Confirm the ssh tunnel is established"
    pass


def create_tunnel(pem_file, server):
    "Create an ssh tunnel."
    pass


def connect_tunnel():
    "Connect to an ssh tunnel"
    pass


def sendlog():
    "Send the log somewhere."
    pass



import os
import regex as re
import serial
import time
import logging
import Simple_Process_REPL.repl as r
import Simple_Process_REPL.utils as u

logger = logging.getLogger()

yaml = u.dump_pkg_yaml("Simple_Process_REPL", "device.yaml")

# format and fill in as you wish.
HelpText = (
    """
Device: - Simple device interaction.  -

If the device path is set, understands how to wait for,
and handshake with a usb device. It can wait for a device path,
and handshake with that device for the purpose of testing.

Device uses this part of the Application state.

%s

The only current use of device is
in conjunction with the particle module which fully populates
the device data structure.

Wait and handshake can work, if only the path to the device is set,
as that is the only thing required.  So, simply set 'device/path'
to your device and everything here should at least try to wait for and
or connect with your device.

The Handshake function

This is a generic function that is a bit more complicated.
It manages an interaction with a device. Everything _handshake_
does is defined in the configuration file. As with everything else,
if anything fails, or doesn't match, an exception is raised.

Here are the steps that _handshake()_ does.

  * Wait for the specified device path to appear.
  * Wait for the _start_string_, match it.
  * Respond with the _response_string_.
  * Look in the output for:
    * fail_regex,
    * done_regex,
    * do_qqc_regex.
  * If fail, raise an exception.
  * if done, exit quietly with true.
  * if do_qqc, then call the do_qqc_function
  and send the return value to the serial device.

  qqc = quelque chose = something.

  In the config the do_qqc_function is set to input-serial,
  as an example. Input-serial prompts for a serial number,
  validates it, and returns it.
  input-serial at the
  _SPR:>_ prompt.

"""
    % yaml
)


def help():
    """Additional SPR specific Help for the device Module."""
    print(HelpText)


def pause():
    "Sleep for configured number of some # of seconds."
    time.sleep(A.get_in_config(["waiting", "pause_time"]))


def wait_for_file(path, timeout):
    """
    Look for a file at path for the given timeout period.
    returns True or False, works for /dev/ttyUSB...
    """
    start = time.time()
    print("Waiting for Path:", path)
    while not os.path.exists(path):
        if time.time() - start >= timeout:
            return False
    return True


def wait():
    """wait for our device to appear."""
    return wait_for_file(
        A.get_in_device("path"), A.get_in_config(["waiting", "timeout"])
    )


def do_qqc(line):
    do_qqc_regex = A.get_in_config(["device", "handshake", "do_qqc_regex"])
    if do_qqc_regex:
        return re.match(do_qqc_regex, line)
    return False


def test_line_fails(line):
    fail_regex = A.get_in_config(["device", "handshake", "fail_regex"])
    result = False
    if re.match(fail_regex, line):
        result = True
        logger.error("handshake failed with: %s" % line)
    return result


def test_done(line):
    done_regex = A.get_in_config(["device", "handshake", "done_regex"])
    result = False
    if re.match(done_regex, line):
        result = True
        logger.info("Test finished with: %s" % line)
    return result


def handshake():
    """
    Handshake with the device after test flash.
    All parameters are set in the configuration.
    Wait for start string, respond with response string.
    Look for fail_regex, done_regex, and do_qqc_regex.
    If fail, raise exception.
    if done, exit quietly with true.
    if do_qqc, then call function and send return to the
    serial device.
    """
    init_string = A.get_in_config(["device", "handshake", "start_string"])
    response_string = A.get_in_config(["device", "handshake", "response_string"])
    do_qqc_func = A.get_in_config(["device", "handshake", "do_qqc_func"])
    baudrate = A.get_in_config(["device", "serial", "baudrate"])
    usb_device = A.get_in_device("path")

    _handshake(usb_device, baudrate, init_string, response_string, do_qqc_func)


def _handshake(usb_device, baudrate, init_string, response_string, do_qqc_func):
    """
    Handshake with the device after test flash.
    All parameters are set in the configuration.
    Wait for start string, respond with response string.
    Look for fail_regex, done_regex, and do_qqc_regex.
    If fail, raise exception.
    if done, exit quietly with true.
    if do_qqc, then call function and send return to the
    serial device.
    """

    result = False  # start with failure, until we get the handshake.

    logger.info("Wait for string: %s" % init_string)
    # timout=None makes this a blocking call that waits.
    with serial.Serial(usb_device, baudrate, timeout=None) as ser:
        got_string = ser.read(size=len(init_string)).decode("utf-8")
        logger.info("Caught: %s" % got_string)

        # send the response and then wait for test results.
        if got_string == init_string:
            logger.info("Sending Response: %s" % response_string)
            ser.write(bytes(response_string, "utf-8"))

            # so far so good, get ready to fail.
            result = True
            logger.info("Waiting for test results.")

            while True:
                line = ser.readline().decode("utf-8")
                logging.info(line)

                if test_line_fails(line):
                    result = False
                    # break
                    ser.close()
                    raise Exception("Fail: %s" % line)

                if do_qqc_func is not None and do_qqc(line):
                    func = r.get_symbol(do_qqc_func)["fn"]
                    response = bytes(func() + "\n", "utf-8")
                    ser.write(response)

                if test_done(line):
                    break

            ser.close()
    return result



import Simple_Process_REPL.core as r
import Simple_Process_REPL.appstate as A

"""
The purpose of this module is to Serve as a liaison between the
Interpreter/REPL and the Application interface layer. This is where
you will add functionalities which you wish to use in your processes.

The Repl module requires two symbol tables, and a dictionary of
defaults, as well as any stateful data the application would like to keep.
In this case it is the _device_ dictionary below.

The symbol table of these functions is defined and passed to
to the repl core module which is reponsible for similar things
as well as parsing the command line, setting up logging, and
executing the process as desired.

The symbol tables could be automated with the use of import. But I have
tried to keep a minimum of magic here. Symbol tables are clear and obvious.
"""

# This is a map which is merged with the application state map.
# Defaults are used by the cli where appropriate.
# device is our particle board, we need it's id, and the
# the path of it's device, ie. /dev/ttyUSB..., The board
# name is boron, photon, whatever. and the last_id just
# in case we need it.

MyState = {}
MyState = A.load_defaults(MyState)

symbols = []
specials = []

# get the default parser for the application and add to it if needed.
parser = None
# parser = r.get_parser()
# parser.add_argument("-f", "--foo", action="store_true", help="set foo")


def init():
    """
    Call into the interpreter/repl with our stuff,
    This starts everything up.
    """
    r.init(symbols, specials, parser)



import argparse
import textwrap

help_prefix = textwrap.dedent(
    """\
        The Simple Process REPL is intended to make it easy
        to define Serialized Processes. It manages the CLI, configuration,
        logging and help, as well as providing multiple ways of running
        the process.

        Define your own processes to interact with anything you like.

        Run interactively in a REPL, run the autoexec once, or in a loop, or
        just run a command from the cli.

        This program is based on a REPL/interpreter. Everything it does
        is defined with it's YAML config file.

        New commands can be defined in the configuration files, and any
        command can be specified to run automatically. The foundational
        commands are python functions. Secondary commands are simply
        lists of other commands, so that more complex processes can be
        created from smaller ones.

        By default the autoexec is set to a 'hello' function.

        Configuration:
        There is a default config file as well as a config file
        option. The default configuration is loaded just before parsing
        the cli. If the config file option is given, it will then be
        loaded on top of the default, and the cli will be re-parsed with
        the new defaults from both configuration files. So all options on
        the command line will win, but also receive their proper defaults.

        After parsing it is possible to load another configuration file.
        It is possible that this should be rearranged. First the
        defaults, then the optional configuration, then reparse with
        the new defaults from both configurations.

        Help:  There are two different help commands available. This one,
        -h, and the internal command, 'help' which gives help for all
        commands available to create processes from. ie. 'spra -h',
        'spra --help', or for the internals, 'spra help'.

        This script assumes there will only be one particle board
        connected at a time. Device detection and targeting would be a
        problem otherwise.

        Interactive mode will loop continually in order to configure
        multiple boards with dialogs. There is a minimized interaction
        mode which prompts only for boards, reporting success or fail,
        for each.

       Each step can also be  manually from the command line.
       Some builtins are:

       * connect wifi
       * python Dialogs
       * cli prompts
       * inspection of state
       * shell commands

       Interactive mode will loop continually with dialogs.
       The default, with no options, is to do everything once in order,
       for one process.
       """
)


help_suffix = textwrap.dedent(
    """\
       Commands can be given after the options in any sequence. Each
       command will run in the order given. Any errors will cause the
       sequence to abort.
        """
)


def define_help_text(prefix, suffix):
    """Add application specific help text"""
    global help_prefix
    global help_suffix

    help_prefix = prefix + "\n" + help_prefix
    help_suffix = suffix + "\n" + help_suffix


def get_argp():
    """Get an argument parser."""
    return argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=help_prefix,
        epilog=help_suffix,
    )


def create_parser(defaults):
    """Create argument parser."""
    parser = get_argp()

    parser.add_argument(
        "--config-file",
        dest="config_file",
        help="Yaml configuration file to load.",
        type=argparse.FileType(mode="r"),
    )

    parser.add_argument(
        "-f",
        "--file",
        dest="file",
        help="File name of spr code to execute",
        action="store",
    )

    parser.add_argument(
        "--logfile",
        default=defaults["logfile"],
        action="store",
        help="logfile to use",
    )

    parser.add_argument(
        "-r",
        "--repl",
        action="store_true",
        help="Start a REPL",
    )

    parser.add_argument(
        "-i",
        "--interactive",
        action="store_true",
        help="Loop, like Awk, do the autoexec or the cli commands in a loop.",
    )

    parser.add_argument("commands", nargs="*")

    return parser


def parse_cli(defaults):
    """parse the command line, provide a dictionary for the defaults."""
    return create_parser(defaults).parse_args()



"""
This is the markdown module for SPR, it can convert markdown to html.
"""
import Simple_Process_REPL.appstate as A
import Simple_Process_REPL.utils as u
import markdown as m
import logging

logger = logging.getLogger()

# Name where we keep our stuff.
Root = "markdown"

yaml = u.dump_pkg_yaml("Simple_Process_REPL", "markdown.yaml")

# format and fill in as you wish.
HelpText = (
    """
markdown: - Converting Markdown to HTML  -


Works with 'With'. so put your markdown in _markdown_.
call md/html,
pop the result where you want it. Probabaly 'html'.

Markdown uses these parts of the Application state.

%s

with /foo

'
markdown: A bunch of markdown.


md/html

pop result html
"""
    % yaml
)


def help():
    """Additional SPR help For the markdown Module."""
    print(HelpText)


def html(markdown):
    """
    convert markdown to html
    """
    return m.markdown(markdown)



import pkgutil

# import logging
import Simple_Process_REPL.logs as logs
import Simple_Process_REPL.repl as r
import Simple_Process_REPL.appstate as A

import Simple_Process_REPL.dialog as D

"""
This file defines the symbol table for the interpreter,
initializes the application, and then runs in the requested manner.

The repl allows for the creation of new function/commands interactively.
Those functions can then be saved as part of the configuration or separately.

The process to be run automatically should be a function in the symbol table
By default, or if not found, the 'hello' function is used.
The autoexec attribute in the configuration names the command(s) to run when
running in a loop or one time with no commands.
"""

logger = logs.setup_logger()

Libs = []


# endless loop with dialog next y/n.
def interactive_loop(commands=None):
    """Execute the autoexec command in an interactive
    loop which reports failures and prompts to do another.
    """
    interactive = A.get_in(["args", "interactive"])
    while interactive is True:
        try:
            do_one(commands)
            if D.continue_to_next():
                interactive = False
        except Exception as e:
            logger.error(e)
            if D.continue_to_next():
                interactive = False
        A.reset_device()


def do_one(commands=None):
    """Execute the default process or the cli commands one time,
    with fail and finish dialogs"""
    try:
        if commands is None:
            A.eval_default_process()
        else:
            r.eval_cmds(commands)

        D.finish()

    except Exception as e:
        logger.error("Process Failed")
        logger.error(e)
        D.failed()
        D.finish()


def do_something():
    """
    Get the startup and shutdown hooks from the config.
    load :
        core.spr
        startup hook
        Load/run a file if given on the cli with -f

    Maybe:
      Start the REPL and run the commands given on the cli.
      Loop over something, the cli commands or the autoexec.
      Do something one time. the cli commands or the autoexec,
    """

    try:
        r.load(pkgutil.get_data(__name__, "core.spr").decode("utf-8").split("\n"))
    except Exception as e:
        logger.error(e)

    commands = A.get_in(["args", "commands"])

    # Each of these should be a list of commands to execute.
    startup_hook = A.get_in(["config", "exec", "hooks", "startup"])

    logger.debug("startup hook: %s", startup_hook)

    if startup_hook:
        r.eval_cmds(startup_hook)

    # Load a file if we got one.
    sprfile = A.get_in(["args", "file"])

    if sprfile:
        try:
            r.load_file(sprfile)
        except Exception as e:
            logger.error(e)

    # Maybe Run the repl.
    if A.get_in(["args", "repl"]):
        r.repl(A.get_in_config(["REPL", "prompt"]), None)

    # or run in a loop
    elif A.get_in(["args", "interactive"]) is True:
        interactive_loop(commands)

    else:
        do_one(commands)


def init(symbols, specials, parser):
    """
    Parse the cli parameters,
    load the default config or the configuration given,
    start logging,
    initialize the symbol tables for the interpreter.
    Finally, run in whatever mode we were told.
    """
    A.init(parser, logger)
    # make the symbol table available in the App State.
    A._set_({"_Root_": r.Root})
    # Need to revisit. Be nice if it worked.
    # A._set_({"_Current_NS_": r.Current_NS})

    # Add fundamental commands to the root level of the interpreter.
    r._import_(
        "Simple_Process_REPL.repl",
        [
            "ns",
            "ls",
            "ns_tree",
            "_help_",
            "_pyhelp_",
            "_def_",
            "_def_path",
            "_import_",
            "partial",
            "in_ns",
            "namespace",
            "load_file",
            "_quit_",
            "eval",
        ],
    )

    # define all the symbols for the things we want to do.
    # r.root_symbols(_symbols, _specials)
    r.root_symbols(symbols, specials)

    do_something()



import sys
import traceback
import pydoc
import pkgutil
import readline
import logging
import regex as re
import os
import atexit
import code
from sys import exit
from inspect import signature, _empty
from Simple_Process_REPL.appstate import (
    merge_yaml_with,
    merge_pkg_yaml,
    _full_with_path,
    select_with,
    push,
    get,
    get_in,
    get_from_path,
    get_in_config,
    get_keys_in,
    show,
)
import Simple_Process_REPL.utils as u

# The same configuration can be stored as instructions in a file read
# by the library with a single call. If myreadline.rc contains:
# # Turn on tab completion
# tab: complete
# # Use vi editing mode instead of emacs
# set editing-mode vi
# the file can be read with read_init_file():

# another alternative.
# Use the tab key for completion
# readline.parse_and_bind('tab: complete')

try:
    readline.read_init_file("spr-readline.rc")
except Exception:
    readline.parse_and_bind("tab: complete")

"""
This is a repl/interpreter with almost no syntax.
It currently recognizes words, strings and numbers.
"""


Root = {}
all_symbols = []  # for the readline completer.
stypes = ["fptr", "voidfptr", "partial", "dolist", "namespace"]
Current_NS = "/"
NS = Root

logger = logging.getLogger()


# yaml = u.load_pkg_yaml("Simple_Process_REPL", "bar_qr.yaml")

# # format and fill in as you wish.
# HelpText=("""
# bar-qr: - Bar and QR Code Scanning, generation, saving and printing.  -

# Recognized codetypes are either 'barcode' and 'QR-code'.

# Encode any value with preset prefix and suffixes as set in the configuration
# into a barcode or QR-code.

# Read a bar or QR code using the webcam. Currently keeps trying until it
# catches it or an escape key is pressed.

# Bar-qr uses this part of the Application state.

# %s

# When encoding the value is retrieved from bar-QR/value.
# When scanning with the webcam the resulting value will be placed in bar-QR/value.

# If working exclusively with bar or QR codes it may be beneficial to define
# partials for some functions, which always require a code type.

# """ % yaml)


# def help():
#     print(HelpText)


def root_symbols(symbols, specials):
    """add symbols to the symbol table as a namespace."""
    global Root
    append_specials(append_symbols(Root, symbols), specials)
    # print(Root.keys())


def append_funcs(st, module, funclist):
    """Import a python module and add the function list to the given symbol table."""
    # get the SPR symbols and stuff from the module.
    lib = __import__(module, globals(), locals(), funclist, 0)

    # logger.info("Append Funcs: %s %d" % (funclist, len(funclist)))
    for fname in list(funclist):
        if type(fname) is not str:
            # logger.info("Not str: %s" % str(fname))
            return
        else:
            f = getattr(lib, fname)

        # figure out if it's a var args [a*], or an [a b c*] varargs.
        # and how many args. Just to be nice and check later.
        # also stash the parameters in case do_fptrs() wants to get fancy.
        signature, nargs, vargs, def_index, parms, sigmap = get_fsig(f)

        name = fname.replace("_", "-")  # cause I don't like hitting shift.

        st[name] = dict(
            fn=f,
            doc=f.__doc__,
            signature=signature,
            pkeys=parms,
            sigmap=sigmap,
            nargs=nargs,
            vargs=vargs,
            def_index=def_index,
            stype="fptr",
        )

    return st


def is_blank_line(line):
    """test if text is blank or not."""
    return re.match(r"^[\s]*$", line)


def inline_yaml(lines):
    """take an iterator and read until the end or two blank lines are encountered.
    Attempt to merge the resulting yaml and merge it into to yaml datastore."""
    yaml = ""
    blank_count = 0

    logger.info("loading inline Yaml:")

    while True:

        try:
            line = lines.__next__()
            logger.info("%s" % line.rstrip())
        except Exception:
            break

        if len(line) == 0 or is_blank_line(line):
            blank_count += 1
        else:
            blank_count = 0

        if blank_count == 2:
            break

        yaml += line + "\n"

    merge_yaml_with(yaml)
    return lines


def load(reader):
    """Load a text reader (hopefully an spr file) into the interpreter."""
    txt = ""
    lines = reader.__iter__()
    line = True
    logger.info("loading SPR:")

    while True:

        try:
            line = lines.__next__()
        except Exception:
            break

        if line and line[0] == "#":
            continue

        if line.strip() == "'":
            inline_yaml(lines)
            txt = ""
            continue

        else:
            if len(line) == 0 or is_blank_line(line):
                if len(txt) > 0:
                    eval_cmd(txt)
                    txt = ""
            else:
                # get rid of newlines so the interpreter
                # sees a continuous line.
                logger.info("%s" % line)
                txt += re.sub("\n", " ", line)

    if len(txt) > 0:
        eval_cmd(txt)


def import_lib_spr(module):
    """Look for spr and yaml files within the python module which are
    named after the module. Load them in the current namespace if found."""
    if re.findall("\.", module):
        root, name = module.split(".")
    else:
        root = module
        name = module
    sprname = name + ".spr"
    yname = name + ".yaml"
    logger.info("\nImport spr %s: %s" % (root, sprname))

    try:
        init = pkgutil.get_data(root, sprname).decode("utf-8").split("\n")
        load(init)
    except Exception:
        pass

    try:
        merge_pkg_yaml(module, yname)
    except Exception:
        pass


def namespace(name, docstr, module, *funclist):
    """
    Create a name space with name. From the python module, import the functions listed.
    Will also import any spr or yaml files found within the module having the
    same name as the module. As per the import_lib_spr function.
    """
    global Root
    logger.info("\nCreating Namespace: %s from Python module: %s" % (name, module))
    ns = {
        "symbols": append_funcs({}, module, *funclist),
        "doc": docstr,
        "name": module,
        "funclist": funclist,
        "stype": "namespace",
    }
    Root[name] = ns
    in_ns(name)
    import_lib_spr(module)


def _import_(module, *funclist):
    """import functions from a python module into the current namespace.
    Then import the module's yaml and spr files if found.

    A complete python-SPR module 'foo' would have a foo.py, foo.spr,
    and possibly a foo.yaml.
    """
    global NS
    if NS == Root:
        append_funcs(NS, module, *funclist)
    else:
        append_funcs(NS["symbols"], module, *funclist)

    import_lib_spr(module)


def ns():
    """Which namespace are we in?"""
    global Current_NS
    global NS
    ns = NS
    if Current_NS == "/":
        print("\nNamepace: %s\n SPR Root Namespace.\n\n" % Current_NS)
    else:
        print(
            "\nNamespace: %s\n   Module: %s\n    %s\n\n"
            % (Current_NS, ns["name"], ns["doc"])
        )


def in_ns(ns=None):
    """Change into a namespace."""
    global Root
    global NS
    global Current_NS
    # logger.info("Root: %s" % Root.keys())

    # logger.info("%s" % Root[ns])
    # for k in Root[ns].keys():
    #     try:
    #         logger.debug("%s" % Root[ns][k].keys())
    #     except Exception:
    #         pass

    if ns is None or ns == "/":
        Current_NS = "/"
        NS = Root
    else:
        Current_NS = ns
        NS = Root[ns]
    logger.info("In Namespace: %s" % Current_NS)


def append_symbols(st, slist):
    """
    Given a list of symbols, add them to the symbol table given.
    each symbol should be in the form of
    ['name', function | str, 'help string']
    """
    for name, function, helptext in slist:
        if isinstance(function, str):
            stype = "dolist"
        else:
            stype = "voidfptr"
        st[name] = dict(fn=function, doc=helptext, stype=stype)
    return st


def find_first_parameter_with_default(parameters):
    """find the index of the first parameter with a default value."""
    res = 0
    count = 0
    for k in parameters.keys():
        # logger.info(parameters[k])
        count += 1
        if parameters[k].default == _empty:
            continue
        else:
            res = count
            break
    return res


def get_fsig_map(parameters):
    """Create a map of parameters with their default values."""
    res = {}
    count = 0
    for k in parameters.keys():
        count += 1
        p = parameters[k]
        v = p.default
        if v == _empty:
            v = None
        res[p.name] = v

    return res


def get_fsig(f):
    if not callable(f):
        return None, 0, False, 0, []

    sig = signature(f)
    parameters = sig.parameters
    nargs = len(parameters)
    pkeys = []
    sigmap = {}
    varargs = False
    def_index = 0
    if nargs > 0:
        pkeys = parameters.keys()
        sigmap = get_fsig_map(parameters)
        last_arg_key = list(pkeys)[nargs - 1]
        last_arg_kind = parameters[last_arg_key].kind.name
        varargs = last_arg_kind in ["VAR_POSITIONAL", "VAR_KEYWORD"]
        def_index = find_first_parameter_with_default(parameters)
    return str(sig), nargs, varargs, def_index, pkeys, sigmap


def append_specials(st, slist):
    """Given a symbol table and a list of specials append them to the
    symbol table. each symbol should be in the form of
    ['name', function | str, nargs, 'help string']
    """
    for name, function, nargs, helptext in slist:
        sig, nargs, vargs, def_index, parms, sigmap = get_fsig(function)
        if sig:
            st[name] = dict(
                fn=function,
                signature=sig,
                pkeys=parms,
                sigmap=sigmap,
                nargs=nargs,
                vargs=vargs,
                def_index=def_index,
                doc=helptext,
                stype="fptr",
            )


def _def_(name, helpstr, commandstr):
    """define a new symbol which is a 'dolist' of other symbols.
    example: def hello-world \"My help text\" ui/msg \"hello world\" """
    stype = "dolist"
    if commandstr[0] == "/":
        stype = "path"
    _def_symbol(name, helpstr, commandstr, stype=stype)


def _def_path(name, helpstr, commandstr):
    """define a new symbol which is a 'path' to another place.
    absolute path;
    example: def mybaz \"My baz path\" /foo/baz

    Relative path to with;
    def mybaz \"My baz path\" foo/baz"""
    stype = "path"

    # if commandstr[0] != "/":
    #     commandstr = _full_with_path(commandstr)

    _def_symbol(name, helpstr, commandstr, stype=stype)


def partial(name, helpstr, commandstr):
    """define a new function from an fptr function which has some
    of it's arguments filled in.

    partial get-bar-code-from 'helptxt' set-in-from barQR value from:

    will create a new function get-bar-code-from that takes a value vector
    just like set-in-from would. Because it is."""
    _def_symbol(name, helpstr, commandstr, stype="partial")


def _def_symbol(name, helpstr, commandstr, stype="dolist"):
    """Define a new symbol in the symbol table of the current namespace."""
    global NS
    global Current_NS
    logging.debug(
        "Define: %s/%s, %s, %s, %s" % (Current_NS, name, commandstr, helpstr, stype)
    )
    s = dict(fn=commandstr, doc=helpstr, stype=stype)
    if NS == Root:
        NS[name] = s
    else:
        NS["symbols"][name] = s


def get_ns(s):
    global Root
    namespace = None
    name = s
    ns = None
    if s.find("/") >= 0:
        ns, name = s.split("/")
        namespace = Root.get(ns)
    return name, namespace


def _get_symbol(s):
    "Quietly, Try to get a symbol's definition."
    global Root
    logger.debug("Getting Symbol: %s" % s)
    n = None
    namespace = None
    # does start with a namespace?
    # This means that / means name space path...
    # traceback.print_stack() # getting called from atom. too much!

    if isinstance(s, str):
        n, namespace = get_ns(s)
        # logger.info("n and n: %s %s" % (n, namespace))
        if namespace is not None:
            symbol = namespace["symbols"].get(n)
        elif Root[s]:
            symbol = Root[s]
        else:
            # get it from the appstate if we can.
            symbol = get(s)
    else:
        symbol = None

    logger.debug("Got Symbol: %s" % symbol)

    return symbol


def get_symbol(s):
    "Try to get a symbol's definition."
    try:
        symbol = _get_symbol(s)
    except Exception:
        symbol = None

    if symbol is None:
        logger.error("Unkown Symbol: %s" % s)

    return symbol


def isstype(v, s):
    if not isinstance(s, list):
        return v["stype"] == s
    else:
        return v["stype"] in s


def all_fptr_help(st):
    """Print help for a python function."""
    for k, v in sorted(st.items()):
        if v["stype"] not in ["fptr", "voidfptr"]:
            continue
        fptr_help(k, v)


def fptr_sig(k, v):
    if "signature" in v.keys():
        return "%s%s" % (k, v["signature"])
    else:
        return k


def fptr_help(k, v, indent=False):
    """Format an fptr symbol's help."""
    sig = fptr_sig(k, v)
    indention = ""
    if indent:
        indention = "    "

    print("\n%s%-20s" % (indention, sig))
    print("\n%sSignature Map: %-40s" % (indention, v["sigmap"]))
    print("%s----------------\n%s%s" % (indention, indention, v["doc"]))


def all_dolist_help(st):
    """give the doc and source for all the compound symbols."""
    if len(st.items()) > 0:
        for k, v in sorted(st.items()):
            dolist_help(k, v)


def dolist_help(k, v):
    """Format an dolist symbol's help."""
    if isstype(v, "dolist") or isstype(v, "partial") or isstype(v, "path"):
        print("\n {0[0]:20}\n--------------\n{0[1]}".format([k, v["doc"]]))
        print("    Type: %s" % v["stype"])

    if isstype(v, "path"):
        print("    Path: [%s]\n" % v["fn"])

    if isstype(v, "fptr"):
        fptr_help(k, v)

    if isstype(v, "partial"):
        print("    Source: [%s]\n" % v["fn"])
        print("  --- Derived from: ---")
        func_name = v["fn"].split(" ")[0]
        s = _get_symbol(func_name)
        fptr_help(func_name, s, indent=True)


def namespace_help(key, ns):
    """Format an namespace symbol's help."""
    print("\n\nNamespace: %s\n   Module: %s\n    %s\n\n" % (key, ns["name"], ns["doc"]))
    try:
        eval_cmd("%s/help" % key)
    except Exception:
        pass
    all_fptr_help(ns["symbols"])
    all_dolist_help(ns["symbols"])


def list_namespace():
    """List the non namespace stuff in the Root namespace."""
    print("\n Root Namespace: ")
    for k, v in sorted(Root.items()):
        if not isstype(v, "namespace"):
            print("   %s" % k)


def ns_tree():
    """List the namespaces and all of their symbols """
    list_namespace()
    print("\n")
    for k, v in sorted(Root.items()):
        if isstype(v, "namespace"):
            printns(k, v)
            for j, z in sorted(v["symbols"].items()):
                print("   %s" % j)


def printns(k, v):
    print(
        "\n%-10s %20s: \n---------------------------------------------\n%s"
        % (k, v["name"], v["doc"])
    )


def printns_syms(ns):
    for k, symbol in sorted(ns["symbols"].items()):
        print_sym(k, symbol)


def get_parent_symbol(symbol):
    """get the symbol this symbol is derived from"""
    func_name = symbol["fn"].split(" ")[0]
    s = _get_symbol(func_name)
    return func_name, _get_symbol(func_name)


def print_sym(k, symbol):
    """Print a symbols signature."""
    sig = fptr_sig(k, symbol)

    if isstype(symbol, "namespace"):
        print("   %-15s -->  %-30s %-30s" % (sig, symbol["name"], symbol["doc"]))

    if isstype(symbol, "partial"):
        parent_name, parent = get_parent_symbol(symbol)
        print("   %-30s  Partial -->  %-20s" % (sig, fptr_sig(parent_name, parent)))

    if isstype(symbol, "dolist"):
        print("   %-30s  Do List -->  %-20s" % (sig, symbol["fn"]))

    if isstype(symbol, "path"):
        print("   %-30s  Path -->  %-20s" % (sig, symbol["fn"]))

    if isstype(symbol, "fptr"):
        print("   %-30s" % sig)


def ls(ns=None):
    """
    List the name spaces, the contents of a namespace.
    Or the contents of the Application state at value vector given
    in the form of a pathname starting with a '/'

    ls - list all symbols in the root namespace.
    ls foo - list all the symbols in the foo namespace
    ls / - show the symbols tree in the root of the Application State
    ls /bar-QR - show the symbol tree from the /bar-QR node of the Application State
    ls /bar-QR/qrcode - show the symbol tree from the /bar-QR node of the Application State
    """
    if ns is None:
        ns = Root
        print("Namespaces:")
        for k, v in sorted(ns.items()):
            if isstype(v, "namespace"):
                print_sym(k, v)
                # printns(k, v)

        print("\nFunctions:")
        for k, v in sorted(ns.items()):
            if not isstype(v, "namespace"):
                print_sym(k, v)
    else:
        if ns[0] == "/":
            if len(ns) == 1:
                keys = get_keys_in()
            else:
                vv = ns[1:].split("/")
                keys = get_keys_in(*vv)

            for k in keys:
                print("    %-30s" % k)
        else:
            s = get_symbol(ns)
            if isstype(s, "namespace"):
                printns(ns, s)
                printns_syms(s)


def help_summary():
    """help"""
    t = """Initially this was a simple alias to wrapped python functions. Such that
    functions would be called sequentially entirely for their side effects which
    would compound into a complete system.

    Start with a configuration and clean system, accumulate information and data,
    do whatever. do it again. or do it once, or do a part of it, etc.

    A different job/application only needs a different configuration file and maybe
    some spr code. Depending on the complexity, maybe an spr/python extension.

    There is an Application datastore which is built from yaml config files
    as well as by spr code within spr extensions.

    An spr extension is not always needed, but provides a mechanism for importing
    yaml data and spr code in conjunction of python code. An spr extension is a standard
    python/pip type module with spr and yaml included. A template project can be created
    with the `new_spr_extension_project` command.

    This language, if you can call it that, has almost no syntax. words, double quoted
    strings and numbers are it. It has no stack, no scope, no variables exactly.
    Because all of the data is initially yaml, There is the idea of paths to data,
    these look like unix paths. ie. /this/is/a/path/to/something.

    There is namespaces, but only one layer deep. Root, '/', has some basic commands,
    and the appstate module imported as 'as' has the rest of the most important commands.

    SPR does have a `with` stack, which can be pointed anywhere in the data store, and it
    understands yaml, so data manipulation within the datastore is very easy.

    The idea is to point at a place in the datastore tree, fill it with what you
    care about, let functions get their parameters from there and put their results
    on the results stack there. Move the results around as needed, push or pop a `with`,
    do another step, etc.  If anything fails, the process fails. Everything is logged.


    yaml can be entered like so.

    `'
     foo:
          bar: 10
          baz: 100
    `

    and will be merged into the application datastore at the location of the current
    `with`.

    as/-with foo
    '
    spam: Hello


    will put a spam entry into foo next to bar and baz.
    """
    print(t)


def helpful_cmds():
    """Print a list of helpful commands"""
    t = """
    The Simple Process REPL is a thin framework of configuration,
    datastore, and REPL.  It can import python modules and use them
    directly. Each module is usually associated with a namespace upon
    import.

    SPR's purpose is to make it easy to create simple application type processes
    which may interact with a user, other devices, or anything else.
    It should be simple to compose processes which are repeatable and consistent
    in behavior. It should also be easy to add new configuration options which can
    also be easily used and over-ridden.

    Python functions will bind to data in the `with` if they find what they want.

    Python functions are imported with namespace or import.
    Partial functions can be defined in spr code.
    Do List functions can be defined in spr code or in the configuration file.
    Datastore chunks can be defined in the configuration files, spr code, or in
    extension yaml files.



    Useful Commands: ls, help, pyhelp, show

    * `help` to get a summary.
    * `help /` to get help for the root namespace.
    * `help <ns>` to get help for a namespace.
    * `help <ns>/<function>` to get help for a function in a namespace.

    * `pyhelp <ns>` to get python help for a namespace.
    * `pyhelp <ns>/<function>` to get python help for a function in a namespace.

    * `ls` to list the contents of the Root, '/',  Namespace.
    * `ls <ns>` to see a summary of a namespace.

    / denotes a path into the application state datastore.
    * `ls /` to list the contents of the Root, '/',  of the application state.
    * `ls /path/to/something` to see a list of keys at that location in the Application State.

    * `show` to browse the Application State Data.
    * `show /path/to/something` to see the contents of the datastore tree at that location.

    By Convention;
    * `show <ns module>` to see the stateful data used by a namespace, if any.
    * `show config <ns module>` to see the configuration data used by the namespace

    * `browse-doc`  To read the README.md in your browser.
    * `view-doc`  To read the README.md in an HTML viewer - does not require internet.
    """
    print(t)


def _help_(args=None):
    """
    This is the SPR help, Additionally there is also the builtin
    python help which can be accessed with the pyhelp command.
    The two help systems show different information, all of which is
    quite useful.

    Help understands SPR symbols and their origins, so it can display
    different and additional information about SPR than the python help.

    Help can be used on namespaces and functions.  All of the following
    Will give help in more and more specific ways.

     * help
     * help /
     * help nw
     * help nw/connect-wifi

    """
    global Root
    application_help()
    sym = args
    print("\nSPR Help\n=============================================")
    if args is None:  # or len(args) == 0:
        helpful_cmds()
        print("=============================================")
        print("All Symbols in the Root / namespace")
        ls()
    elif args == "/":
        all_fptr_help(Root)
        all_dolist_help(Root)
        print("=============================================")

    else:
        # for sym in args:
        # logger.info("Sym in args: " % (sym, args))
        if sym[0] == "/":
            sym = sym[1:]
        s = _get_symbol(sym)
        if isstype(s, "namespace"):
            namespace_help(sym, s)
        if isstype(s, ["dolist", "partial", "path"]):
            dolist_help(sym, s)
        if isstype(s, ["fptr", "voidfptr"]):
            fptr_help(sym, s)


def _pyhelp_(args=None):
    """
    This is the interactive python help, Additionally there is also the builtin
    SPR help which can be accessed with the help command.
    The two help systems show different information, all of which is
    quite useful.

    pyHelp understands Python modules and doc strings and can therefore
    provide nicely formated python level help for all SPR functionality.

    It does also know how to show the documentation for the underlying
    python functions which are used by SPR's partials.

    pyHelp can be used on namespaces and functions.  All of the following
    Will give help in more and more specific ways.

     * pyhelp
     * pyhelp nw
     * pyhelp nw/connect-wifi
    """

    global Root
    application_help()

    sym = args
    helpful_cmds()
    if args is None:  # or len(args) == 0:
        print("\nSPR Help\n=============================================")
        # all_fptr_help(Root)
        # all_dolist_help(Root)
        helpful_cmds()
        print("=============================================")
        print("All Symbols in the Root / namespace")
        ls()

    else:
        # for sym in args:
        # logger.info("Sym in args: " % (sym, args))
        s = _get_symbol(sym)
        if isstype(s, "namespace"):
            pydoc.help(s["name"])
        if isstype(s, "partial"):
            n, s = get_parent_symbol(s)
            pydoc.help(s["fn"])
        if isstype(s, ["dolist", "path"]):
            dolist_help(sym, s)
        if isstype(s, ["fptr", "voidfptr"]):
            pydoc.help(s["fn"])


def application_help():
    """Print help as defined by the function set in ['exec', 'help']
    in the config. This is a help function as defined by the
    application layer, which is specific to the functionality
    that we are interfacing with.
    """
    helpfn = get_in_config(["exec", "help"])
    if helpfn:
        eval_cmd(get_in_config(["exec", "help"]))


def all_symbol_names():
    """Return a list of all symbol names. for readline, prompt completion."""
    res = []
    res.extend(list(Root.keys()))
    for k, v in sorted(Root.items()):
        if isstype(v, "namespace"):
            res.extend(v["symbols"].keys())

    return res


def isa_fptr(cmd):
    "Quietly determine if a symbol is a fptr with args"
    try:
        sym = _get_symbol(cmd)
        if not isstype(sym, "fptr"):
            sym = None
    except Exception:
        return None
    else:
        return sym


def isa_path(cmd):
    "Quietly determine if a symbol is a path"
    try:
        sym = _get_symbol(cmd)
        if not isstype(sym, "path"):
            sym = None
    except Exception:
        return None
    else:
        return sym


def get_specials_with_narg(st, nargs):
    """Get a list of specials which have number of args defined."""
    swn = []
    for k, v in st.items():
        if isstype(v, "fptr") and v["nargs"] == nargs:
            swn.append(k)
    return swn


def get_sig_map(symbol, path):
    fptr = isa_fptr(symbol)
    if fptr:
        sigmap = fptr["sigmap"]
    if path is None:
        set(_get_with_path(), sigmap)


def expand(commands):
    """Expand the arguments into their values as needed.
    So far this is just paths. With this, we have 'variables'.
    Path symbols expand to their value.

    Keeping it simple, one layer expansion, no recursion. It's enough
    to give dangerous power.
    """
    res = []
    for symbol in commands:
        path = isa_path(symbol)
        if path:
            logger.debug("Expand: %s" % path["fn"])
            v = get(path["fn"])
            if not isinstance(v, str):
                res += [v]
            else:
                res += v.split(" ")
        else:
            res += [symbol]

    logger.debug("Expand: %s" % res)
    return res


def resolve_vars(commands, start_index=1):
    """Big cheat, just gonna try to expand variables and see what happens.
    So yea, it works. Sorta. Too well. too much. Need a more real interpreter.
    on verra.
    """

    # if len(commands) > start_index:
    #     cmds = commands[:start_index]
    #     for symbol in commands[start_index:]:
    #         if isinstance(symbol, str):
    #             v = get(symbol)
    #             if v:
    #                 cmds += [v]
    #             cmds += [symbol]
    #         else:
    #             cmds += [symbol]
    #     commands = cmds

    return commands


# I think this function is too complicated. could be simpler. but it is
# working at the moment.
def do_fptrs(commands):
    """Call the fptr functions with their argument lists.
    Validates the number of arguments if possible. nargs=-1 is varible args.
    returns true if it was a special, false if not, even if it fails.

    This get's the first shot. if the first token isn't an fptr, then it goes
    on to be evaluated elsewhere.
    """

    result = None
    command = commands[0]

    # logging.debug("do fptrs: %s" % command)
    # logging.debug("do fptrs: %s" % command)
    # logging.debug("full: %s" % commands)
    # # logging.debug("fptr: %s" % str(fptr))

    fptr = isa_fptr(command)

    if fptr is None:
        return False

    # logging.info("keys: %s" % str(special.keys()))
    # # logging.debug("nargs: %s" % str(special['nargs']))

    # Such a mess. This needs to be fixed so it uses args and kwargs,
    #  goes hand in hand with making with work.

    # logger.info("DO_FPTR %s" % fptr)
    fn = fptr["fn"]
    fnargs = fptr["nargs"]
    vargs = fptr["vargs"]
    def_index = fptr["def_index"]
    sig = fptr["signature"]
    nargs = len(commands) - 1
    pkeys = fptr["pkeys"]

    # logger.info("do-fptrs: %d, %s" % (nargs, commands))

    # logger.info("args: %s" % args)

    # logger.info("Fptr: %s\nNargs: %d\nVargs: %d\nSig: %s" % (fptr, nargs, vargs, sig))
    # logger.info(
    #     "Command: %s - %s %s\n argcount: %d prms: %d varargs: %d"
    #     % (command, fn, sig, fnargs, nargs, vargs)
    # )

    # oy. je n'aime pas des exceptions
    if command == "-def-" or command == "partial":
        # commands = resolve_vars(commands, start_index=2)
        commandstr = " ".join(commands[3:])
        fn(commands[1], commands[2], commandstr)
        return True

    # oy. je n'aime pas des exceptions
    if command == "as/set":
        # noop ? - yes
        # commands = resolve_vars(commands, start_index=2)
        fn(commands[1], commands[2:])
        return True

    # noop ? - yes
    # commands = resolve_vars(commands, start_index=1)

    try:
        if fnargs == 0 and nargs == 0:
            result = fn()

        elif vargs:

            args = commands[1:fnargs]
            args += [commands[fnargs:]]
            result = fn(*args)

        # elif nargs <= fnargs and nargs >= def_index - 1:
        else:

            args = dict(zip(pkeys, commands[1:]))
            with_vars = select_with(list(pkeys)[nargs:])

            try:
                args = dict(zip(pkeys, commands[1:]))
                if with_vars and nargs < fnargs:
                    args.update(with_vars)

            except Exception as e:
                print(e)

            result = fn(**args)

            if result is not None:
                push("results", result)

    except Exception as e:
        logger.error(commands)
        print(e)
        # traceback.print_exception(*sys.exc_info())

        return True


def atom(token):
    'Numbers become numbers; "..." string; otherwise Symbol/cmd.'
    if token[0] == '"':
        # to get rid of the quotes, but that doesn't serve us.
        # return token[1:-1].encode('utf_8').decode('unicode_escape')
        return token.encode("utf_8").decode("unicode_escape")
    try:
        return int(token)
    except ValueError:
        try:
            return float(token)
        except ValueError:
            try:
                return complex(token.replace("i", "j", 1))
            except ValueError:
                # A parameter less function in the Root symbol table
                # if valid_command_ck(token):
                return token
            # else:
            #    raise SyntaxError("Unknown command %s" % token)


def tokenize(line):
    "Return the next token, reading new text into line buffer if needed."
    tregex = r"""\s*("(?:[\\].|[^\\"])*"|[^\s(")]*)(.*)"""
    tokens = []
    while True:
        if line == "":
            break
        token, line = re.match(tregex, line).groups()
        if token != "":
            tokens.append(token)
    logging.debug("tokenize: %s" % tokens)
    return tokens


def read(tokens):
    "Read an expression and return a list of tokens to be evaluated."
    commands = []
    for token in tokens:
        commands.append(atom(token))
    logging.debug("read %s" % commands)
    return commands


def parse(commandstr):
    "Tokenize a command string and parse it."
    logging.debug("parse %s" % commandstr)
    return read(tokenize(commandstr))


def eval_symbol(s):
    "Execute a function in the symbol table."
    logger.debug("eval symbol: %s" % s)

    try:
        symbol = get_symbol(s)
        if symbol is not None:
            function = symbol["fn"]

            # its a string of symbols.
            if isinstance(function, str):
                eval_list(parse(function))
            else:
                function()
    except Exception:
        pass


def eval_list(commands):
    """Execute a list from the symbol table.
    It's either a partial, a command with
    arguments or a list of commands.
    """
    logger.debug("eval list: %s" % commands)

    # we have function pointers sort-of by having this here.
    # if you put a symbol name for something in the yaml
    # and a path symbol that points at it....
    commands = expand(commands[0:])

    # logger.info(commands)
    # if it's a partial expand it.
    first_sym = _get_symbol(commands[0])

    if first_sym is None:
        logger.error("Unkown Symbol: %s" % commands[0])
        return False

    if isstype(first_sym, "partial"):
        logger.debug("Partial: %s --> %s" % (commands[0], commands))
        eval_list(parse(first_sym["fn"]) + commands[1:])
        return

    if do_fptrs(commands) is False:
        for cmd in commands:
            eval_symbol(cmd)


def eval_cmd(commandstr):
    """Evaluate a commandstr."""
    eval_list(parse(commandstr))


def eval_cmds(commands):
    """Evaluate a command list."""
    for c in commands:
        eval_list(parse(c))


def eval(path):
    """Evaluate a list of symbols, from a path, if what is held there is
    a string, then evaluate it.
    If it is a list evaluate each entry in turn.

    ```
    set mymessage cli/msg "hello there"

    eval mymessage
    ```

    An example using 'with' to construct a list of commands.

    ```
    with my/func/of/stuff

    '
    - cli/msg hello
    - cli/msg goodbye

    pop-with
    eval /my/func/of/stuff

    def myfunc "My function that does more than one thing." eval /my/func/of/stuff
    ```
    """
    v = get(path)
    if isinstance(v, list):
        eval_cmds(v)
    else:
        eval_cmd(v)


def repl_message():
    """Give a welcome message."""
    print("Welcome to the REPL.")
    print("Type 'help' for help and 'quit' to exit.")
    helpful_cmds()


def make_completer(vocabulary):
    def custom_complete(text, state):
        # None is returned for the end of the completion session.
        results = [x for x in vocabulary if x.startswith(text)] + [None]
        # A space is added to the completion since the Python readline doesn't
        # do this on its own. When a word is fully completed we want to mimic
        # the default readline library behavior of adding a space after it.
        return results[state] + " "

    return custom_complete


class HistoryConsole(code.InteractiveConsole):
    def __init__(
        self,
        locals=None,
        filename="<console>",
        histfile=os.path.expanduser("~/.spr-history"),
    ):
        code.InteractiveConsole.__init__(self, locals, filename)
        self.init_history(histfile)

    def init_history(self, histfile):
        readline.parse_and_bind("tab: complete")
        if hasattr(readline, "read_history_file"):
            try:
                readline.read_history_file(histfile)
            except FileNotFoundError:
                pass
            atexit.register(self.save_history, histfile)

    def save_history(self, histfile):
        readline.set_history_length(1000)
        readline.write_history_file(histfile)


def load_file(filename):
    """load an SPR file into the application."""
    logger.info("Loading SPR file: %s" % filename)
    with open(filename, "r") as reader:
        load(reader)


def _quit_():
    shutdown_hook = get_in(["config", "exec", "hooks", "shutdown"])
    if shutdown_hook:
        eval_cmds(shutdown_hook)
    exit()


def yaml_parse():
    yaml = ""
    while True:
        try:
            line = input("YAML...>")
            if len(line):
                yaml += line + "\n"
            else:
                break
        except Exception as e:
            logger.error(e)
    merge_yaml_with(yaml)
    return


def repl(prompt="SPR:> ", init=None):
    """
    An input loop. nothing fancy. It does have history and
    completion at least.
    """

    all_symbols = all_symbol_names()
    readline.parse_and_bind("tab: complete")
    readline.set_completer(make_completer(all_symbols))
    HistoryConsole()

    if init is not None:
        eval_cmds(init)

    repl_message()

    while True:
        try:
            print("")
            line = input(prompt)
            if line and line[0] == "#":
                continue
            if line == "'":
                yaml_parse()
            else:
                if len(line):
                    eval_list(parse(line))

        except Exception as e:
            print(e)



import Simple_Process_REPL.appstate as A
import subprocess
import os
import time
import logging

logger = logging.getLogger()
AS = A.AS

# format and fill in as you wish.
HelpText = """
Subcmd: - SPR's subprocess interface, ie. shell.  -

Use the sub process command library to execute system commands.

Roughly equivalent, but not, to python's os.system().

https://docs.python.org/3/library/subprocess.html

"""


def help():
    """Additional SPR specific Help for the App State Module."""
    print(HelpText)


def rm(filename):
    """remove file at path."""
    os.remove(filename)


def environ():
    """Show the shell environment"""
    logger.info(os.environ)


def sleep(s):
    """Sleep for x seconds"""
    time.sleep(s)


def mk_cmd(cmd, prefix=""):
    """
    Put together a command list for subprocess.run.
    simplistic, split on spaces to make the list of command pieces.
    """
    if prefix:
        command = [prefix]
    else:
        command = []
    command.extend(cmd.split(" "))
    logger.debug(command)
    return command


def do_cmd(command, shell=False):
    """run a sub command, read and return it's output."""
    if shell is True:
        command = " ".join(command)

    logging.debug("do_cmd: %s" % command)

    res = subprocess.run(
        command, shell=shell, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )

    stderr = res.stderr.decode("utf-8")
    stdout = res.stdout.decode("utf-8")

    logger.debug("Return Code: %d" % res.returncode)
    logger.info(stdout)
    logger.error(stderr)

    if res.returncode:
        raise Exception("stdout: %s\n stderr: %s\n" % (stdout, stderr))

    return stdout


def do(*commands):
    """A varargs wrapper for do_cmd using subprocess to run shell commands."""
    do_cmd(*commands)



# Create the barcode stickers and QR code stickers
import barcode
from barcode.writer import ImageWriter
import qrcode
from PIL import Image, ImageDraw, ImageFont, ImageOps
import Simple_Process_REPL.utils as u
import os
import logging

import cv2
from pyzbar import pyzbar

logger = logging.getLogger()

BarCodeType = "barcode"
QRCodeType = "QR_code"

Root = "bar-QR"

yaml = u.dump_pkg_yaml("Simple_Process_REPL", "bar_qr.yaml")
spr = u.load_pkg_resource("Simple_Process_REPL", "bar_qr.spr")

# format and fill in as you wish.
HelpText = """
bar-qr: - Bar and QR Code Scanning, generation, saving and printing.  -

Recognized codetypes are either 'barcode' and 'QR-code'.

Encode any value with preset prefix and suffixes as set in the configuration
into a barcode or QR-code.

Read a bar or QR code using the webcam. Currently keeps trying until it
catches it or an escape key is pressed.

Bar-qr uses this part of the Application state.

%s

When encoding the value is retrieved from bar-QR/value.
When scanning with the webcam the resulting value will be placed in bar-QR/value.

If working exclusively with bar or QR codes it may be beneficial to define
partials for some functions, which always require a code type.

bar-qr has the following spr code.

%s

""" % (
    yaml,
    spr,
)


def help():
    """Additional SPR help For the Bar and QR code Module."""
    print(HelpText)


def gen(value, codetype=BarCodeType):
    """Generate a bar or QR code for the value.
    returns: the generated code
    """
    try:
        if codetype == BarCodeType:
            code = create_bar_code(value)
        elif codetype == QRCodeType:
            code = create_qr_code(value)

    except Exception as e:
        logger.error(e)

    return code


def save(value, code, codetype=BarCodeType):
    """Save the codetype, 'barcode/QR_code',
    for value with generated code, code.

    Write code for value to a file with a nice name generated
    from the value and type.

    returns: filename
    """
    if codetype == BarCodeType:
        fname = get_bc_filename(value)
        save_barcode(code, fname)
        # because the extension is automatic on the save.
        fname = "%s%s" % (fname, ".png")
    elif codetype == QRCodeType:
        fname = get_qr_filename(value)
        save_qr_code(code, fname)

    return fname


def write(value, codetype=BarCodeType):
    """Generate and write a bar or QR code for
    value into a file which is appropriately named.

    Returns: filename."""
    c = gen(value, codetype)
    filename = save(value, c, codetype)
    return filename


def pad_num(n):
    """make sure its a number and is the minimum length, pad from left with 0s."""
    try:
        min = int(A.get_in_config(["bcqr_minimum_length"]))
    except Exception:
        min = 10
    fmt = "%%.%dd" % min
    return fmt % int(n)


def build_code(s, prefix, suffix):
    """turn a number into what we need for a barcode."""
    if prefix is not None:
        code = "%s%s" % (prefix, pad_num(s))
    else:
        code = pad_num(s)

    if suffix is not None:
        code = "%s%s" % (code, suffix)

    return code


def num_2_barcode(s):
    """turn a number into what we need for a barcode."""
    return build_code(
        s,
        A.get_in_config(["barcode", "prefix"]),
        A.get_in_config(["barcode", "suffix"]),
    )


def num_2_qrcode(s):
    return build_code(
        s,
        A.get_in_config(["QR_code", "prefix"]),
        A.get_in_config(["QR_code", "suffix"]),
    )


def save_barcode(bc, filename):
    # options = [module_height = 8, text_distance = 2]
    path = A.get_in_config(["barcode", "save_path"])
    options = A.get_in_config(["barcode", "save_options"])
    bc.save(filename, options)


def get_bc_filename(s):
    """generate a name for a barcode file"""
    suffix = A.get_in_config(["barcode", "filename_suffix"])
    path = A.get_in_config(["barcode", "save_path"])
    os.makedirs(path, exist_ok=True)
    return os.path.join(path, "%s%s" % (s, suffix))


def get_qr_filename(s):
    """generate a name for a QR code file"""
    suffix = A.get_in_config(["QR_code", "filename_suffix"])
    path = A.get_in_config(["QR_code", "save_path"])
    os.makedirs(path, exist_ok=True)
    return os.path.join(path, "%s%s%s" % (s, suffix, ".png"))


def save_qr_code(qrc, filename):
    qrc.save(filename)


def create_bar_code(s):
    """Create a barcode from a string"""
    return barcode.get("code128", num_2_barcode(s), writer=ImageWriter())


def create_qr_code(s):
    """Create a QR code from a string"""
    qrfont = A.get_in_config(["QR_code", "font"])
    qrfont_size = A.get_in_config(["QR_code", "font_size"])

    qrCode = num_2_qrcode(s)
    qr = qrcode.QRCode(
        version=2,
        error_correction=qrcode.constants.ERROR_CORRECT_L,
        box_size=4,
        border=0,
    )

    qr.add_data(qrCode)
    qr.make(fit=True)

    img = qr.make_image(fill_color="black", back_color="white")
    backColor = "rgb(255, 255, 255)"
    # Left Top Right Bottom
    bimg = ImageOps.expand(img, border=(30, 0, 30, 40), fill=backColor)

    draw = ImageDraw.Draw(bimg)
    font = ImageFont.truetype(qrfont, size=qrfont_size)
    (x, y) = (34, 116)
    color = "rgb(0, 0, 0)"
    draw.text((x, y), qrCode, fill=color, font=font)

    return bimg


def makeFailSticker(reason, code):
    """make a failure sticker.  Currently Unused"""
    img = Image.new("L", (200, 100), 255)
    draw = ImageDraw.Draw(img)
    font = ImageFont.truetype("DejaVuSans.ttf", size=24)
    (x, y) = (1, 8)
    color = "rgb(0, 0, 0)"
    draw.text((x, y), "TEST FAIL", fill=color, font=font)
    font = ImageFont.truetype("DejaVuSans.ttf", size=14)
    (x, y) = (1, 33)
    case = reason + ": " + str(code)
    draw.text((x, y), case, fill=color, font=font)
    return img


def read_barcodes(frame):
    success = False
    barcodes = pyzbar.decode(frame)
    for barcode in barcodes:
        x, y, w, h = barcode.rect
        barcode_info = barcode.data.decode("utf-8")
        cv2.rectangle(frame, (x, y), (x + w, y + h), (0, 255, 0), 2)
        font = cv2.FONT_HERSHEY_DUPLEX
        cv2.putText(frame, barcode_info, (x + 6, y - 6), font, 2.0, (255, 255, 255), 1)

        A.set_in([Root, "value", barcode_info])
        logging.info("Read Code: %s" % barcode_info)
        success = True

    return frame, success


# Need to take a closer look at this stuff. It's not very nice.
# read barcodes did show a nice green rectangle and the code but
# now it goes away as soon as it matches, so no help.
# maybe a way to get the rectangle back?  I don't know.
# needs some expermenting I think.
def read_barcode_from_camera():
    """Read a Bar or QR code from the camera."""
    camera = cv2.VideoCapture(0)
    ret = True
    while ret:
        ret, frame = camera.read()
        frame, res = read_barcodes(frame)
        # frame, res = read_code(frame)
        if res:
            break
        cv2.imshow("Barcode/QR code reader", frame)
        if cv2.waitKey(1) & 0xFF == 27:
            break
    camera.release()
    cv2.destroyAllWindows()
    return frame


def read_code(image):
    # initialize the cv2 QRCode detector
    res = False
    detector = cv2.QRCodeDetector()
    # detect and decode
    data, vertices_array, binary_qrcode = detector.detectAndDecode(image)
    # if there is a QR code
    # print the data
    if vertices_array is not None:
        A.set_in([Root, "value", data])
        res = True

    return image, res


def video_to_frames(video, path_output_dir):
    # extract frames from a video and save to directory as 'x.png' where
    # x is the frame index
    vidcap = cv2.VideoCapture(video)
    count = 0
    while vidcap.isOpened():
        success, image = vidcap.read()
        if success:
            cv2.imwrite(os.path.join(path_output_dir, "%d.png") % count, image)
            count += 1
        else:
            break
    cv2.destroyAllWindows()
    vidcap.release()


# video_to_frames('../somepath/myvid.mp4', '../somepath/out')



import Simple_Process_REPL.appstate as A
import Simple_Process_REPL.utils as u
from Simple_Process_REPL.dialog import yes_no, inputbox, msg
import Simple_Process_REPL.particle as P
import regex as re
import logging
import os


"""
The purpose of this module is to Serve as a liaison between the
Interpreter/REPL and the device interface layer, the particle-cli,
which in this case is defined in particle.py
"""

# This is a map which is merged with the application state map.
# Defaults are used by the cli where appropriate.
# device is our particle board, we need it's id, and the
# the path of it's device, ie. /dev/ttyUSB..., The board
# name is boron, photon, whatever. and the last_id just
# in case we need it.

logger = logging.getLogger()

yaml = u.dump_pkg_yaml("Simple_Process_REPL", "particle_main.yaml")
spr = u.load_pkg_resource("Simple_Process_REPL", "particle_main.spr")

# format and fill in as you wish.
HelpText = """
particle: - A particle.io device interface.  -

Interact with particle boards.  list, identify, claim, flash, test, etc.

Particle uses these parts of the Application state

%s

Particle Defines This SPR code.

%s

When working with a Particle.io board, the first step is to 'get'
the id and device. There are many commands here which need the usb device
or the device id. So the first step is always a `get`

A 'get' will populate the device in the Application
state with the currently connected device.

After that, it is possible to continue working with that device
to update it, set it's setup bit, flash it, handshake with it etc.

If a repeatable process is desired, simply create one by defining more
complex commands.

Working  in the REPL with particle boards is much nicer than the particle-cli.

""" % (
    yaml,
    spr,
)


def help():
    """Additional SPR specific Help for the Particle Module."""
    print(HelpText)


def see_images():
    """Ask if the test images are seen, fail if not."""
    yno_msg = A.get_in_config(["dialogs", "images_seen"])
    if not yes_no(yno_msg):
        raise Exception("Failure: Test images not Seen.")


def input_serial():
    """
    Function called by test handshake, Asks if images are seen,
    then presents a dialog loop to receive an 8 digit serial number
    to be returned back to the handshake and sent to the device.
    """
    see_images()

    msg = A.get_in_config(["dialogs", "input_serial"])
    while True:
        code, res = inputbox(
            msg, title=A.get_in_config(["dialogs", "title"]), height=10, width=50
        )
        if re.match(r"^\d{8}$", res):
            yno_msg = "%s : %s" % (
                A.get_in_config(["dialogs", "serial_is_correct"]),
                res,
            )
            if yes_no(yno_msg):
                break
        else:
            msg(A.get_in_config(["dialogs", "serial_must"]))

            os.system("clear")
            logger.info("Serial Number Entered is: %s" % res)
            A.set_in(["device", "serial", res])
        return res


def flash_image():
    "Flash the flash image"
    P.flash(A.get_in_config(["particle", "images", "flash"]))


def flash_tinker():
    "Flash the tinker image"
    P.flash(A.get_in_config(["particle", "images", "tinker"]))


def flash_test():
    "Flash the test image"
    P.flash(A.get_in_config(["particle", "images", "test"]))


def product_add():
    """ product device add product id - associate the device wit a product"""
    P.product_add(A.get_in_config(["particle", "product"]), A.get_in_device("id"))


def claim():
    "Claim the device, cloud claim"
    P.claim(A.get_in_device("id"))


def add():
    "Claim the device, device add"
    P.add(A.get_in_device("id"))


def release():
    "Release the claim on a device."
    P.release(A.get_in_device("id"))


def name(name):
    """Name/Rename the current device."""
    P.name(A.get_in_device("id"), name)


def name_from(varpath):
    """Name/Rename the device from a variable in the Application state."""
    n = A.get_in(varpath)
    logger.info("Naming device: %s" % n)
    name(n)


def cloud_status():
    "Check to see if the device is connected to the cloud"
    P.cloud_status(A.get_in_device("id"))


def reset_usb():
    "Reset the usb device."
    P.reset_usb(A.get_in_device("id"))


def list_usb_w_timeout():
    "Do a serial list repeatedly for timeout period"
    P.list_usb_w_timeout(A.get_in_config(["waiting", "timeout"]))


def get_usb_and_id():
    """
    Retrieve and set the USB device, the board name,  and the device id.
    Uses 'particle serial list' in a timeout loop. This is required
    for most things. Wait and handshake, use the usb device,
    and the id is needed by many things.
    """
    # PB['usb_device'], PB['device_id'] = P.get_usb_and_id()
    path, name, id = P.get_w_timeout(A.get_in_config(["waiting", "timeout"]))
    A._set_({"device": {"path": path, "name": name, "id": id}})


def wait_for_plist():
    """Wait for particle serial list to succeed with timeout, doesn't
    really work."""
    # we don't care about the results, just to wait.
    P.get_w_timeout(A.get_in_config(["waiting", "timeout"]))


def archive_log():
    "Move the current logfile to one named after the current device."
    archive_log("%s.log" % A.get_in_device("id"))


def do_pcmd(cmds):
    """Call the particle cli, with something."""
    P.do_pcmd(" ".join(cmds))


def particle_help():
    """A function to provide additional Application specific help."""
    print(
        """\n\n Particle Process help:\n
Particle help can be accessed directly at any time with the 'particle-help'
command.\n
When working with a Particle.io board, the first step is to 'get'
the id and device. There are many commands here which need the usb device
or the device id.  With that, the basic commands for the process we have
been intending would be;

'start, setup, claim, testit, flash, and archive_log'.

Note that 'start', 'setup', and 'testit', are user defined commands which are
actually lists of other commands. These are defined in the configuration
file.  Help lists the source for these commands.
'start' is actually; 'dialog-start get wait identify'

The 'continue to next' and 'device failed' dialog windows are built in to
the interactive loop. Any additional prompts, such as 'dialog-start' or
'dialog-test' can be added to the process  with their commands.
and prompt texts are defined in the configuration, which can be
seen with 'show config'.

When a process is deemed good, the autoexec can be set to it, and that
will be what runs automatically, in the process loop, or as a oneshot,
if no commands are given on the cli.\n\n """
    )


# nicer names in this name space. not pretty, but easy.
dfu = P.dfu_mode
listen = P.listen
list = P.list_usb
identify = P.identify
inspect = P.inspect
login = P.login
logout = P.logout
update = P.update
setup_done = P.set_setup_bit
_doctor = P.doctor
_flash = P.flash
pcmd = P.do_pcmd



import logging
import sys


"""
This is a mess. Nothing is really working properly. Nothing is going
to the log file, and the simpler code just doesn't work. For sure,
to get logging for everything, do not give a name to getLogger so
the root logger will come back.

I'm not really sure why start_logging()
doesn't work but setup_logger() does.

So more work to do here.

We are at least getting log info to stdout all the way
down into the particle.py code.

"""

message_fmt = "[%(asctime)s] %(levelname)-8s %(name)-12s %(message)s"

stdout_format = "%(message)s"

# message_fmt = '[%(asctime)s] %(levelname)-8s %(name)-12s %(message)s'
# message_fmt = '[%(levelname)s] %(name)s %(asctime)s - %(message)s'
message_fmt = "[%(name)s] %(asctime)s - %(message)s"


def validate_loglevel(loglevel):
    nloglevel = getattr(logging, loglevel.upper(), None)
    if not isinstance(nloglevel, int):
        raise ValueError("Invalid log level: %s" % nloglevel)
    return nloglevel


# this one was not working but maybe it will now. I'd like
# to study it.
def start_logging(loglevel):
    """Start up logging with 2 handlers. One for the logfile and
    on for stdout. Set the log level the same to all loggers."""

    nloglevel = validate_loglevel(loglevel)

    # file_handler = logging.FileHandler(logfile)
    stdout_handler = logging.StreamHandler(sys.stdout)
    handlers = [stdout_handler]
    logger = logging.getLogger()

    logging.basicConfig(
        encoding="utf-8", format=message_fmt, level=nloglevel, handlers=handlers
    )

    return logger()


def set_level(logger, lvl):
    "Set the level for the logger."
    logger.setLevel(validate_loglevel(lvl))


def log_lvl(lvl):
    """Change the logging level."""
    set_level(logging.getLogger(), lvl)


def add_file_handler(logger, loglevel, filename):
    """
    Takes a logger, a string loglevel, and a filename,
    to create and add a file handler to a logger.
    """

    nloglevel = validate_loglevel(loglevel)
    formatter = logging.Formatter(message_fmt)
    fh = logging.FileHandler(filename, mode="w", encoding="utf-8")
    # if we set these, then changing the root level has no effect.
    # fh.setLevel(nloglevel)
    fh.setFormatter(formatter)
    logger.addHandler(fh)


def setup_logger():
    """"setup a root level logger with the stdout handler"""
    log = logging.getLogger()
    out_hdlr = logging.StreamHandler(sys.stdout)
    out_hdlr.setFormatter(logging.Formatter(stdout_format))
    # out_hdlr.setLevel(logging.INFO)
    log.addHandler(out_hdlr)
    log.setLevel(logging.INFO)

    log.info("logger setup to info")

    return log


info = logging.info
warning = logging.warning
error = logging.error
critical = logging.critical
debug = logging.debug
level = log_lvl

# format and fill in as you wish.
HelpText = """
log: - This is the builtin logging of SPR -

Control the logging level and send messages to the log.
"""


def help():
    """Additional SPR specific Help for the logs Module."""
    print(HelpText)



import pathlib

# python 3.9 is required.
try:
    # Use setup() from setuptools(/distribute) if available
    from setuptools import setup
except ImportError:
    from distutils.core import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

from Simple_Process_REPL import __version__

setup(
    name="Simple_Process_REPL",
    version=__version__,
    # Your name & email here
    long_description=README,
    long_description_content_type="text/markdown",
    description="An extensible application framework with REPL for creating processes with side effects.",
    url="https://github.com/EricGebhart/Simple_Process_REPL",
    author="Eric Gebhart",
    author_email="e.a.gebhart@gmail.com",
    packages=["Simple_Process_REPL"],
    include_package_data=True,
    package_data={"": ["*.yaml", "*.spr", "*.txt"]},
    scripts=[],
    license="MIT",
    entry_points={"console_scripts": ["SPR=Simple_Process_REPL.__main__:main"]},
    install_requires=[
        "pyserial",
        "pythondialog",
        "pyYAML",
        "regex",
        "python-barcode",
        "qrcode",
        "Pillow",
        "opencv-python",
        "numpy",
        "Image",
        "Pyzbar",
        "markdown"
        # "brother_ql",
    ],
)



import pkgutil
import os
import logging
import yaml

logger = logging.getLogger()

# yaml = A.load_pkg_yaml("Simple_Process_REPL", "appstate.yaml")

# # format and fill in as you wish.
# HelpText = (
#     """
# utils: - SPR's utils, yaml and stuff.  -

# Everything needed to load, save, copy, set, and merge data.
# Yaml files and merging etc.

# Loading of files from python modules.

# """
#     % yaml
# )


# def help():
#     print(HelpText)


def merge(a, b, path=None, update=True):
    """nice solution from stack overflow, non-destructive merge.
    http://stackoverflow.com/questions/7204805/python-dictionaries-of-dictionaries-merge
    """
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            elif isinstance(a[key], list) and isinstance(b[key], list):
                for idx, val in enumerate(b[key]):
                    if len(a[key]):
                        a[key][idx] = merge(
                            a[key][idx],
                            b[key][idx],
                            path + [str(key), str(idx)],
                            update=update,
                        )
                    else:
                        a[key] = b[key]

            elif update:
                a[key] = b[key]
            else:
                raise Exception("Conflict at %s" % ".".join(path + [str(key)]))
        else:
            a[key] = b[key]
    return a


def make_dict(keys):
    """Create a dictionary tree with a value from a list.
    ie. make_dict(["foo", "bar", value]) => {foo: {bar: value}}
    """
    d = {}
    v = None
    for x in reversed(keys):
        if v is None:
            v = x
        else:
            d = {x: v}
            v = d
    return v


def save_yaml_file(filename, dictionary):
    "Write a dictionary as yaml to a file"
    with open(filename, "w") as f:
        yaml.dump(dictionary, f)


def load_yaml_file(filename):
    "load a dictionary from a yaml file"
    someyaml = None
    if os.path.isfile(filename):
        logger.info("Loading YAML: %s" % filename)
        with open(filename) as f:
            someyaml = yaml.load(f, Loader=yaml.SafeLoader)
    return someyaml


def load_pkg_resource(package, filename):
    """Load a file from a python package, and return it."""
    return pkgutil.get_data(package, filename).decode("utf-8")


def load_pkg_yaml(package, yamlname):
    """load a configuration file from a package."""
    try:
        someyaml = yaml.load(
            load_pkg_resource(package, yamlname), Loader=yaml.SafeLoader
        )
        logger.info("Loading YAML from Module: %s: %s" % (package, yamlname))
        return someyaml
    except FileNotFoundError:
        pass
    except Exception as e:
        print(e)


def dump_pkg_yaml(package, yamlname):
    """load a yaml file from a python package and dump it."""
    return yaml.dump(load_pkg_yaml(package, yamlname))



"""
This is basic functionality to view a chunk of html or a url

So far, this is really basic usage of webbrowser and pywebview.

   https://docs.python.org/3/library/webbrowser.html

   https://pywebview.flowrl.com
"""
import Simple_Process_REPL.appstate as A
import Simple_Process_REPL.utils as u
import os
import logging

import webbrowser
import webview
import time

logger = logging.getLogger()

# Name where we keep our stuff.
Root = "web"

yaml = u.dump_pkg_yaml("Simple_Process_REPL", "webview.yaml")

# format and fill in as you wish.

HelpText = (
    """
Web: - View HTML from a url or a chunk.  -

Web uses these parts of the Application state.

%s

Web View will show the HTML, instead of the url if both are given.

The easiest way to use them is by using a with as it is not currently
possible to give a None, or use keyword arguments otherwise.

 with /foo

 '
 url : http://ericgebhart.com
 title: My viewer title.
 open_tab: True

 web/view

 web/browse

 pop-with

 with /foo web/view

 with /foo web/browse


The viewer is very basic.
The browser only understands urls, and if you want a new tab or window.

"""
    % yaml
)


def help():
    """Additional SPR specific Help for the Web Module."""
    print(HelpText)


# Get something out of my part of the config.
# A.get_in_config([Root, "foo"])

# Put something into my part of the Appstate.
# A.set_in([Root, "foo" "bar" 10])

# Get something out of my part of the Appstate.
# A.get_in([Root, "foo" "bar"])


def view(html=None, url=None, title=""):
    """
    The html wins over url if both are given.

    Uses PyWebViewer

    example: with /readme web/view

    """
    webview.create_window(
        title,
        html=html,
        url=url,
        confirm_close=True,
    )
    webview.start()


def browse(url, open_tab=None):
    """Browse the url, Will try to use your default browser.

    If open tab is not set, use the setting
    in /config/browser/open-tab.

    If not true, a new browser window will open.
    """
    if open_tab is None:
        open_tab = A.get_in_config([Root, "browser", "open-tab"])

    if open_tab:
        webbrowser.open_new_tab(url)
    else:
        webbrowser.open_new(url)

    webview.start()
